-- transformation of boolean expressions

module Data.Expr.Proposition.Transform
       ( module Data.Expr.Proposition.Transform
       , module Data.Transform.Maybe
       )
       where

import Data.Expr.Proposition.Constr
import Data.Expr.Proposition.Eval
import Data.Expr.Proposition.Types

import Data.Transform.Maybe

-- ----------------------------------------
--
-- combinators for expression transformation

-- apply a transformation to all nodes of an expr top down
--
-- TODO: substitute by more general transTopDown and transBottomUp

applyTTopDown :: Transformer Expr -> Expr -> Expr
applyTTopDown tf e
  = case applyTrans tf e of
     (Unary  op e1)
       -> Unary op  (recurse e1)
     (Binary op e1 e2)
       -> Binary op (recurse e1) (recurse e2)
     e' -> e'
  where
    recurse = applyTTopDown tf

-- here we need a real transformer
applyTBottomUp :: Transformer Expr -> Expr -> Expr
applyTBottomUp tf e
  = case e of
     (Unary  op e1)
       -> applyTrans tf (Unary  op (recurse e1))
     (Binary op e1 e2)
       -> applyTrans tf (Binary op (recurse e1)
                                   (recurse e2))
     _ -> applyTrans tf e
  where
    recurse = applyTBottomUp tf

-- ----------------------------------------
--
-- special transformations for boolean expressions

-- partial evaluation and
-- .==. substituted by .&&., .||. and not

transEquiv :: Transformer Expr
transEquiv (Binary Equiv (Lit v1) e2)
  |     v1 = success e2
  | not v1 = success (not' e2)

transEquiv (Binary Equiv e1 (Lit v2))
  |     v2 = success e1
  | not v2 = success (not' e1)

transEquiv (Binary Equiv e1 e2)
  = success (e1 .&&. e2 .||. not' e1 .&&. not' e2)

transEquiv _
  = failure

-- ./=. substituted by .&&., .||. and not

transXor :: Transformer Expr
transXor (Binary Xor e1 e2)
  = success (e1 .&&. not' e2 .||. not' e1 .&&. e2)
transXor _
  = failure

-- .=>. substituted by .||. and not

transImpl :: Transformer Expr
transImpl (Binary Impl e1 e2)
  = success (not' e1 .||. e2)
transImpl _
  = failure

transAnd :: Transformer Expr
transAnd (Binary And e1@(Lit v1) e2)
  |     v1 = success e2   -- true  is a 1
  | not v1 = success e1   -- false is a 0

transAnd (Binary And e1 e2@(Lit v2))
  |     v2 = success e1   -- true  is a 1
  | not v2 = success e2   -- false is a 0

transAnd (Binary And e1 (Unary Not e2))
  | e1 == e2 = success false

transAnd (Binary And (Unary Not e1) e2)
  | e1 == e2 = success false

transAnd _
  = failure

transOr :: Transformer Expr
transOr (Binary Or e1@(Lit v1) e2)
  |     v1 = success e1  -- true  is a 0
  | not v1 = success e2  -- false is a 1

transOr (Binary And e1 e2@(Lit v2))
  |     v2 = success e2  -- true  is a 0
  | not v2 = success e1  -- false is a 1

transOr _
  = failure

-- double negation removed
-- and or de Morgan applied

transNot :: Transformer Expr
transNot (Unary Not (Unary Not e1))
  = success e1

transNot (Unary Not (Binary And e1 e2))
  = success $ not' e1 .||. not' e2

transNot (Unary Not (Binary Or e1 e2))
  = success $ not' e1 .&&. not' e2

transNot _
  = failure

-- partial evaluation

transLit :: Transformer Expr
transLit (Unary op (Lit v))
  = success (Lit (mf1 op v))

transLit (Binary op (Lit v1) (Lit v2))
  = success (Lit (mf2 op v1 v2))

transLit _
  = failure

-- ----------------------------------------
