-- | general transformation combinators
--
-- The Maybe result indicates whether a transformation
-- could be performed: @Nothing@ not applicable,
-- else just the result
--
-- More advanced @Transformer@ types would e.g. log all
-- the transformation steps performed
-- in a complex transformation.
--
-- This could be done by combining the Maybe monad
-- with a writer monad.

module Data.Transform.Maybe where

import Control.Applicative ((<$>))

type Transformer a = a -> Maybe a

success :: a -> Maybe a
success = Just

failure :: Maybe a
failure = Nothing

-- choice, orElse is an alias for mplus from MonadPlus
orElse :: Maybe a -> Maybe a -> Maybe a
orElse x1@(Just _) _  = x1
orElse _           x2 = x2

-- apply transformation to an expr
-- if the transformation could not be applied
-- the expression remains unchanged

applyTrans :: Transformer a -> (a -> a)
applyTrans tf e
  = case tf e of
     Nothing -> e
     Just e' -> e'

always :: Transformer a -> Transformer a
always tf e
    = case tf e of
        Nothing -> Just e
        e'      -> e'

-- ----------------------------------------
--
-- transformers

-- don't transform anything

transNothing :: Transformer a
transNothing = const failure

-- try two transfomers sequentially,
-- if at least on can be applied, the transformation succeeds

transSeq :: Transformer a -> Transformer a -> Transformer a
transSeq tf1 tf2 e
  = ( do e' <- tf1 e
         always tf2 e'
    )
    `orElse` tf2 e

transSequence :: [Transformer a] -> Transformer a
transSequence = foldr transSeq transNothing

-- apply to transformations in parallel

transPar :: Transformer a -> Transformer b -> Transformer (a, b)
transPar tf1 tf2 (e1, e2)
  = ( do e1' <- te1                      -- tf1 with success
         e2' <- te2 `orElse` success e2
         success (e1', e2')
    )
    `orElse`                              -- tf2 with success
    ( do e2' <- te2
         success (e1, e2')
    )
    where
      te1 = tf1 e1
      te2 = tf2 e2

transPar' :: Transformer a -> Transformer (a, a)
transPar' tf = transPar tf tf

transParallel :: Transformer a -> Transformer [a]
transParallel _tf []
    = success []
transParallel tf (x : xs)
    = (\ (x', xs') -> (x' : xs')) <$>
      transPar tf (transParallel tf) (x, xs)

-- try a 1. transformation, if not applicable
-- try a 2. one

transAlt :: Transformer a -> Transformer a -> Transformer a
transAlt tf1 tf2 e
  = tf1 e `orElse` tf2 e

transAlternatives :: [Transformer a] -> Transformer a
transAlternatives = foldr transAlt transNothing

-- repeatedly apply a transformation to an expr
-- caution: no cyclic transformations please

transRep :: Transformer a -> Transformer a
transRep tf e
  = do e' <- tf e
       always (transRep tf) e'

-- ----------------------------------------

type Bijection a b = (a -> b, b -> a)

transWrapped :: Bijection a b -> Transformer b -> Transformer a
transWrapped (deconstruct, reconstruct) tf x
    = reconstruct <$> (tf . deconstruct) x

-- ----------------------------------------
